"""Generator for random Icelanders"""
from __future__ import unicode_literals
from builtins import object

import random
import datetime
import json
import os

from kennitala import Kennitala

from .name_scraper import NameScraper
from .address import Address

__author__ = "7oi"
__version__ = "0.3.1"
__license__ = 'MIT'


class Icelander(object):
    """The Icelander generator
    Class that contains methods to generate random icelanders and even dump them to a file
    """

    genders = ['female', 'male']
    names = {
        'female': [],
        'male': [],
    }
    address_generator = Address()

    def __init__(self):
        file_path = os.path.dirname(__file__)
        with open(os.path.join(file_path, NameScraper.DATA_FOLDER, NameScraper.NAMES_FILE), 'r') as names_file:
            self.names = json.loads(names_file.read())

    def get_first_name(self, gender='female'):
        """Generate first name for person

        Keyword Arguments:
            gender {str} -- Gender of person (default: {'female'})

        Returns:
            str -- First name for person
        """

        return random.choice(self.names[gender])[0]

    def get_last_name(self, gender='female', parent=None):
        """Generate last name for person

        Keyword Arguments:
            gender {str} -- Gender of person. (default: {'female'})
            parent {dict} -- Parent dict to derive last name from. If None a random parent name is chosen. (default: {None})

        Returns:
            str -- Last name for person
        """
        if parent is None:
            parent_name = random.choice(self.names[random.choice(self.genders)])[1]
        else:
            parent_name = [
                name
                for _nm, name in self.names[parent.get('gender')]
                if _nm == parent.get('firstname')
            ][0]
        if gender == 'male':
            return '{}son'.format(parent_name)
        return '{}dóttir'.format(parent_name)

    def get_random_person(self, gender=None, year=None, parent=None, address=None):
        """Get random person as a dict

        Keyword Arguments:
            gender {str} -- Gender of person. If None a random gender is selected. (default: {None})
            year {int} -- Birth year of person. If None a random year is selected. (default: {None})
            parent {dict} -- Parent to derive last name from. If None a random last name is selected. (default: {None})
            address {dict} -- An address to associate person with. If None a random address is selected. (default: {None})

        Returns:
            dict -- A dict with attributes for person {gender, firstname, lastname, ssn}
        """

        if year is not None:
            # Select random date in given year
            start_date = datetime.date(year, 1, 1).toordinal()
            end_date = datetime.date(year, 12, 31).toordinal()
            random_date = datetime.date.fromordinal(random.randint(start_date, end_date))
            ssn = Kennitala.generate(random_date)
        else:
            ssn = Kennitala.random()

        if gender is None:
            gender = random.choice(self.genders)

        if address is None:
            address = self.address_generator.get_random_address()
        return {
            'gender': gender,
            'firstname': self.get_first_name(gender),
            'lastname': self.get_last_name(gender, parent=parent),
            'ssn': ssn,
            'address': address
        }

    def get_random_people(self, num_people=1, gender=None, year=None, address=None):
        """Gets a list of random people

        Keyword Arguments:
            num_people {int} -- Number of people to generate. (default: {1})
            gender {str} -- Gender for people. If None a random gender is selected for each person (default: {None})
            year {int} -- Birth year for people. If None a random year is selected for each person (default: {None})

        Returns:
            list -- List of person dicts
        """

        return [
            self.get_random_person(gender=gender, year=year, address=address)
            for _i in range(num_people)
        ]

    def get_random_household(self, size=5, parents=2, place=None, po_code=None):
        """Gets random household

        Keyword Arguments:
            size {int} -- Size of household. (default: {5})
            parents {int} -- Number of parents (default: {2})
            place {str} -- Name of city/town to place them in. If None then selected randomly. (default: {None})
            po_code {str} -- Post code to place them in. If None then selected randomly. (default: {None})

        Returns:
            list -- List of household members
        """
        address = self.address_generator.get_random_address(place=place, po_code=po_code)
        curr_year = datetime.date.today().year
        ret = []
        parent = None
        parent_years_range = range(curr_year - 60, curr_year - 20)
        if parents > 0:
            parent_years = [random.choice(parent_years_range) for _i in range(parents)]
            ret.extend([
                self.get_random_person(year=year, address=address)
                for year in parent_years
            ])
            parent = ret[0] if ret else None
        else:
            # If no parents, give some random years anyway
            parent_years = [random.choice(parent_years_range) for _i in range(3)]

        num_kids = size - parents
        if num_kids > 0:
            kids_years = [
                random.choice(range(random.choice(parent_years) + 18, curr_year))
                for _i in range(num_kids)
            ]

            ret.extend([
                self.get_random_person(year=year, address=address, parent=parent)
                for year in kids_years
            ])
        return ret

    def dump_random_people_to_file(self, filename='random_people_dump.json', num_people=1, gender=None, year=None):
        """Dumps generated people to a json file

        Arguments:
            filename {str} -- Filename to dump to (default: {'random_people_dump.json'})
        """
        with open(filename, 'w') as outfile:
            json.dump(self.get_random_people(num_people, gender, year), outfile, indent=2)



"""The scraper of names from wikipedia"""
from __future__ import print_function
from __future__ import unicode_literals
from builtins import object

import json
import os
import requests

from lxml import html


class NameScraper(object):
    """
    NameScraper is made for scraping the icelandic names from wikipedia and dumping them
    into a json file for use in IcelanderGenerator
    """

    WIKI_FORMAT = 'https://is.wikipedia.org/{}'
    WIKI_FEMALES = 'wiki/Listi_yfir_%C3%ADslensk_eiginn%C3%B6fn_kvenmanna'
    WIKI_MALES = 'wiki/Listi_yfir_íslensk_eiginnöfn_karlmanna'
    WIKI_NAME_XPATH = '//div[@id="mw-content-text"]/div/ul/li/a'
    WIKI_GENETIVE_XPATH = '//*[contains(text(), "Eignarfall")]'
    DATA_FOLDER = 'data'
    NAMES_FILE = 'icelandic_names.json'
    NON_GENETIVE_NAMES_FILE = 'non_genetive_names.json'
    names = {
        'female': [],
        'male': []
    }
    non_genetive_names = {
        'female': [],
        'male': []
    }

    def get_name(self, item):
        """Get name for xpath item

        Arguments:
            item {object} -- Xpath object for name

        Returns:
            tuple | None -- Returns tuple with (name, genetive name) or None if genetive name is not found
        """

        url = item.attrib.get('href')
        name = item.text
        genetive_name = name
        name_page = requests.get(self.WIKI_FORMAT.format(url))
        name_page_tree = html.fromstring(name_page.content)
        try:
            # The xpath should give the <a> element that has the genetive label
            label_element = name_page_tree.xpath(self.WIKI_GENETIVE_XPATH)[0]
            # The element we want is the next cell in the table, thus we need to get the parents sibling
            genetive_name = label_element.getparent().getparent().getchildren()[1].text
        except IndexError:
            return None
        return (name, genetive_name)

    def scrape_wiki_for_names(self):
        """Scrape icelandic wikipedia page for names.
        Should only be used for updating icelandic_names.json file from time to time. As it's a scraper
        that visits multiple wikipedia pages it will take a while to do it's thing.
        """
        for gender, path in [('female', self.WIKI_FEMALES), ('male', self.WIKI_MALES)]:
            page = requests.get(self.WIKI_FORMAT.format(path))
            tree = html.fromstring(page.content)
            items_in_name_xpath = tree.xpath(self.WIKI_NAME_XPATH)[0:-2]
            for item in items_in_name_xpath:
                name = self.get_name(item)
                if name is None:
                    name = item.text
                    print('No genetive version found. Not adding {} to names'.format(name))
                    self.non_genetive_names[gender].append(name)
                    continue
                print('Adding: ({}, {}) to {} names'.format(name[0], name[1], gender))
                self.names[gender].append(name)

    def save_names_to_file(self):
        """Saves names to a json file
        """
        file_path = os.path.dirname(__file__)
        with open(os.path.join(file_path, self.DATA_FOLDER, self.NAMES_FILE), 'w') as outfile:
            json.dump(self.names, outfile, indent=2)

        with open(os.path.join(file_path, self.DATA_FOLDER, self.NON_GENETIVE_NAMES_FILE), 'w') as outfile:
            json.dump(self.non_genetive_names, outfile, indent=2)

    def update_names(self):
        """Scrapes all the names and dumps them to the json file
        """

        self.scrape_wiki_for_names()
        self.save_names_to_file()



from os import path
from setuptools import setup


this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), 'r') as f:
    long_description = f.read()


setup(
    name='icelander_generator',
    packages=['icelander_generator'], # this must be the same as the name above
    version='0.3.1',
    description='A utility to generate random Icelanders',
    author='7oi',
    author_email='7oi@7oi.is',
    long_description=long_description,
    long_description_content_type="text/markdown",
    license='MIT',
    url='https://github.com/7oi/IcelanderGenerator',
    install_requires=[
        'lxml',
        'requests',
        'kennitala',
        'future'
    ],
    include_package_data=True,
    keywords=['tests', 'generator'],
    classifiers = [],
)



"""Generator for Icelandic addresses"""
from __future__ import unicode_literals
from builtins import object

import csv
import random
from os import path


class Address(object):
    """Generate icelandic addresses
    Generates random icelandic addresses from existing street names and po codes.
    House numbers are randomly generated, so they can be non-existent.
    """
    DATA_FOLDER = 'data'
    PO_CODES_FILE = 'po_codes.tsv'
    STREETS_FILE = 'streets.tsv'
    streets = []

    def __init__(self):
        self.import_streets()

    def import_streets(self):
        """Imports streets with correct place names and postcodes from tsv files
        """

        curr_dir = path.dirname(__file__)
        # Read postal codes from tsv file
        with open(path.join(curr_dir, self.DATA_FOLDER, self.PO_CODES_FILE), 'rt', encoding='utf-8') as poc_file:
            postal_codes = {
                po_code: place
                for po_code, place in csv.reader(poc_file, delimiter='\t')
            }

        # Read street names from csv file
        with open(path.join(curr_dir, self.DATA_FOLDER, self.STREETS_FILE), 'rt', encoding='utf-8') as streets:
            for po_code, street in csv.reader(streets, delimiter='\t'):
                place = postal_codes.get(po_code)
                self.streets.append((street, po_code, place))

    def get_random_address(self, place=None, po_code=None, max_num=99):
        """Generates a random street address

        Keyword Arguments:
            place {str} -- Place where to restrict address to. If None no restrictions apply. (default: {None})
            po_code {str} -- Postal code to restrict address to. If None no restrictions apply. (default: {None})
            max_num {int} -- Maximum house numbers on street. (default: {99})

        Returns:
            dict -- {nr, street, po_code, place}
        """

        house_nr = random.randint(1, max_num)
        choice = None

        if po_code is not None:
            try:
                choice = random.choice([street for street in self.streets if street[1] == po_code])
            except IndexError:
                pass
        if place is not None and choice is None:
            try:
                choice = random.choice([street for street in self.streets if street[2] == place])
            except IndexError:
                pass
        if choice is None:
            choice = random.choice(self.streets)
        return {
            'nr': house_nr,
            'street': choice[0],
            'po_code': choice[1],
            'place': choice[2]
        }
